import falcon
from ar_sandbox import Sandbox
from io import BytesIO
import sys
import cv2
import numpy as np
import numpy.ma as npma
from esrimap.esrimap import EsriAscii
from PIL import Image
import requests
import time
import math
import json
import asyncio
import websockets

NORMALIZATION_BETWEEN = (0.8, 1.2)
LO_NORM = NORMALIZATION_BETWEEN[0]
HI_NORM = NORMALIZATION_BETWEEN[1]
SHOW_LINES = True
USE_HARD_COLORS = False
LINE_COLORS = [255, 255, 255]

COLORS_FULL_SPECTRUM_SOFT = [
    (0.2, (0, 0, 255)),
    (.25, (0, 255, 255)),
    (.3, (0, 255, 0)),
    (.35, (255, 255, 0)),
    (.45, (255, 128, 0)),
    (.5, (255, 0, 0)),
    (.75, (255, 0, 128)),
    (.9, (255, 0, 255)),
    (1, (255, 255, 255))
]

COLORS_FULL_SPECTRUM_HARD = [
    (0.2, (0, 0, 255)),
    (0.24999999, (0, 0, 255)),
    (.25, (0, 255, 255)),
    (.29999999, (0, 255, 255)),
    (.3, (0, 255, 0)),
    (.34999999, (0, 255, 0)),
    (.35, (255, 255, 0)),
    (.44999999, (255, 255, 0)),
    (.45, (255, 128, 0)),
    (.49999999, (255, 128, 0)),
    (.5, (255, 0, 0)),
    (.74999999, (255, 0, 0)),
    (.75, (255, 0, 128)),
    (.89999999, (255, 0, 128)),
    (.9, (255, 0, 255)),
    (.9999999, (255, 0, 255)),
    (1, (255, 255, 255)),
]
COLORS = COLORS_FULL_SPECTRUM_SOFT
#
POINTS = np.array([c[0] for c in COLORS_FULL_SPECTRUM_SOFT])
COLOR_R = [c[1][0] for c in COLORS_FULL_SPECTRUM_SOFT]
COLOR_G = [c[1][1] for c in COLORS_FULL_SPECTRUM_SOFT]
COLOR_B = [c[1][2] for c in COLORS_FULL_SPECTRUM_SOFT]
#
POINTS_HARD = np.array([c[0] for c in COLORS_FULL_SPECTRUM_HARD])
COLOR_R_HARD = [c[1][0] for c in COLORS_FULL_SPECTRUM_HARD]
COLOR_G_HARD = [c[1][1] for c in COLORS_FULL_SPECTRUM_HARD]
COLOR_B_HARD = [c[1][2] for c in COLORS_FULL_SPECTRUM_HARD]

COLORMAP = None
COLORMAP_HARD = None

SOCKET_URI = "ws://localhost:8765/?client_id=lib&operation="

def try_and_ignore(ignore=Exception, default_value=0.0):
    """
    Silently ignore specific exception from a decorated function
    Return a default value if exception occur.
    """
    def decorate(func):
        def _decorated(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ignore:
                return default_value
        return _decorated
    return decorate


parse_int = try_and_ignore(ValueError, 0)(int)
parse_float = try_and_ignore(ValueError, 0.0)(float)

api_content_esri = ""
api_content_aruco = ""
api_content_unity_png = None

class FakeSandbox:
    """Placeholder for testing purposes"""
    def __init__(self, refresh=250):
        self.on_frame = None
        self.refresh = refresh

    def init(self):
        pass

    def start(self, *args):
        cap = cv2.VideoCapture(0)
        while cap.isOpened():
            ret, fr = cap.read()
            frame = cv2.resize(fr, (588, 432))
            mat = np.loadtxt("depth")

            mat2 = frame
            if self.on_frame is not None:
                frame = self.on_frame(mat, mat2)

            cv2.imshow("frame", frame)

            k = cv2.waitKey(self.refresh)
            if k == 27:
                break
        cv2.destroyAllWindows()
        cap.release()


def get_color(depth, hard_colors=False) -> np.ndarray:
    """
    Given a depth matrix between 0-1, return a color from the colormap

    Arguments:
        depth : Normalized 1 channel numpy matrix
        hard_colors: Whether to use hard colormap or not
    Return:
        A 3 channel BGR matrix of the same size as depth_matrix.
    """
    depth_matrix = depth.copy()

    if hard_colors:
        pts = COLORMAP_HARD[0]
        r = COLORMAP_HARD[1]
        g = COLORMAP_HARD[2]
        b = COLORMAP_HARD[3]
    else:
        pts = COLORMAP[0]
        r = COLORMAP[1]
        g = COLORMAP[2]
        b = COLORMAP[3]

    r_val = np.interp(depth_matrix, pts, r)
    g_val = np.interp(depth_matrix, pts, g)
    b_val = np.interp(depth_matrix, pts, b)
    res = np.dstack((b_val, g_val, r_val)).astype(np.uint8)
    return res


def display_level(frame):
    out_frame = frame.copy()
    # out_frame = normalize_array_between(out_frame, -4, 1)
    # out_frame = np.clip(out_frame, 0, 1)
    print(out_frame, out_frame.min(), out_frame.max())
    out_frame = get_color(out_frame)
    return out_frame


def send_request(body, url="http://localhost:8000/map"):
    try:
        requests.post(url, json=body)
    except Exception as e:
        print(e)


# This is the older method that was used to transmit arucos over the network
def build_req_body(map, tag_list):
    arucos = []
    for t in tag_list:
        print(t)
        (x, y), id, r = t
        (x, y) = (parse_int(x), parse_int(y))
        id = parse_int(id)
        r = parse_float(r)
        arucos.append({
            "tag_id": id,
            "position": {
                "x": x,
                "y": y
            },
            "rotation": r
        })
    body = {
        "esri": map.to_ascii(),
        "arucos": arucos
    }
    return body


# This new method is basically a copy pasted version of the one above without the map
# Maybe both of them are not needed anymore
def build_arucos_json(tag_list):
    arucos = []
    for t in tag_list:
        print(t)
        (x, y), id, r = t
        (x, y) = (parse_int(x), parse_int(y))
        id = parse_int(id)
        r = parse_float(r)
        arucos.append({
            "tag_id": id,
            "position": {
                "x": x,
                "y": y
            },
            "rotation": r
        })
    body = {
        "arucos": arucos
    }
    return body

def normalize_array(array):
    return (array - array.min())/(array.max() - array.min())


def fix_norm(array, min, max):
    return (array - min) / (max - min)


def normalize_array_between(array, lower_bound, higher_bound):
    norm = normalize_array(array)
    diff = higher_bound - lower_bound
    return (norm * diff) + lower_bound


def make_esri_and_send(depth, color, *args):
    global api_content_esri, api_content_aruco, api_content_unity_png

    if COLORMAP is None or COLORMAP_HARD is None:
        gen_global_default_cmaps()
    # Normalization step
    depth = np.clip(depth, LO_NORM, HI_NORM)  # Clip aberrants values
    depth = np.interp(depth, [LO_NORM, HI_NORM], [0, 1])  # Normalize array
    depth = 1 - depth  # Invert result, making 1 the highest, 0 the lowest.

    # Create level frame, based on depth values
    if USE_HARD_COLORS:
        levels = get_color(depth, hard_colors=True)
    else:
        levels = get_color(depth, hard_colors=False)

    if SHOW_LINES:  # Enable lines between region
        if USE_HARD_COLORS:
            lvl_image = levels
        else:
            lvl_image = get_color(depth, hard_colors=True)
        canny = cv2.Canny(lvl_image, 100, 200)
        can = normalize_array(canny)
        mask = npma.masked_where(can == 1, can)
        levels[mask.mask] = LINE_COLORS

    # Aruco detection
    grayscale_frame = cv2.cvtColor(color, cv2.COLOR_BGR2GRAY)
    
    _, (corn, ids, rej), centers, rots = detect_aruco(grayscale_frame) #TODO

    # Create map and tag dictionary
    esri_map = EsriAscii.from_ndarray(depth)
    detected = build_tag_list(ids, centers, rots) #TODO

    # Send it toward api
    #send_request(build_req_body(esri_map, detected), url=DEFAULT_URL)
    #TODO
    #send_request(build_req_body(esri_map, detected), url=dest)
    #send_request(build_req_body(esri_map, []), url=dest)
    #send_request(build_req_body(esri_map, detected), url=DEFAULT_URL2)

    
    # May locally save the esri_map & arucos
    # esri_map.to_file("depth_map.asc")
    api_content_esri = esri_map.to_ascii()
    api_content_aruco = build_arucos_json(detected)
    # tags = build_arucos_json(detected)

    # with open("aruco_tags.json", "w") as fp:
    #     json.dump(tags, fp)

    # asyncio.run(push_height())
    asyncio.run(pull_unity_frame())

    if api_content_unity_png is not None:
        cv2.resize(api_content_unity_png, (levels.shape[0], levels.shape[1]))
        im_rgb = cv2.cvtColor(api_content_unity_png, cv2.COLOR_BGR2RGB)
        return im_rgb

    # Return value will be displayed on the beamer
    return levels


def build_tag_list(tags_ids, tags_centers, tags_rotation):
    detected = []
    for i, c in enumerate(tags_centers):
        detected.append((c, tags_ids[i][0], tags_rotation[i]))
    return detected


def get_center(corner):
    c0, _, c2, _ = corner[0]
    return int(c0[0] + (c2[0]-c0[0]) // 2), int(c0[1] + (c2[1]-c0[1]) // 2)

# TODO def detect_aruco(a, b=None, c=None): return

def detect_aruco(frame,
                 aruco_dict=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_50),
                 aruco_params=cv2.aruco.DetectorParameters()):
    
    #(corners, ids, rejected) = cv2.aruco.detectMarkers(frame, aruco_dict, parameters=aruco_params)
    detector = cv2.aruco.ArucoDetector(aruco_dict, aruco_params)
    (corners, ids, rejected) = detector.detectMarkers(frame)

    centers = []
    rotations = []
    for c in range(len(corners)):
        try:
            ctr = get_center(corners[c])
            top_left = corners[c][0][0]
            centers.append(get_center(corners[c]))
            (w1, w2) = [-1, -1]
            (v1, v2) = top_left - ctr
            rotation_in_degree = math.degrees(math.atan2(w2 * v1 - w1 * v2, w1 * v1 + w2 * v2))
            rotations.append(round(rotation_in_degree, 2))
        except Exception as e:
            print(e)
    return None, (corners, ids, rejected), centers, rotations


def save_matrices(depth, frame, *args):
    np.savetxt("depth", depth)
    for i in range(frame.shape[2]):
        np.savetxt("frame" + "RGB"[i], frame[:, :, i].astype(np.uint8))
    Image.fromarray(frame).save("frame.png")
    return frame


def create_hard_colors(color_array):
    hard_colors = []
    epsilon = 1e-8
    for i, c in enumerate(color_array[:-1]):
        hard_colors.append(c)
        col = ((color_array[i+1][0] - epsilon), c[1])
        hard_colors.append(col)
    hard_colors.append(color_array[-1])
    return hard_colors


def create_colormap(color_array):
    pts = np.array([c[0] for c in color_array])
    r_chan = [c[1][0] for c in color_array]
    g_chan = [c[1][1] for c in color_array]
    b_chan = [c[1][2] for c in color_array]
    return pts, r_chan, g_chan, b_chan


def gen_global_default_cmaps():
    global COLORMAP, COLORMAP_HARD
    print("Generating default colormaps")
    COLORMAP = create_colormap(color_array=COLORS_FULL_SPECTRUM_SOFT)
    COLORMAP_HARD = create_colormap(color_array=create_hard_colors(color_array=COLORS_FULL_SPECTRUM_SOFT))


def hex_to_rgb(value):
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))


def rgb_to_hex(rgb):
    r, g, b = [int(i) for i in rgb.split(",")]
    rgb = (r, g, b)
    return '#%02x%02x%02x' % rgb

class ApiUnity:
    def on_get(self, req, resp):
        resp.media = api_content_esri

    # def on_post(self, req, resp):
    #     global api_content_unity_png
    #     image_data = req.bounded_stream.read()
    #     api_content_unity_png = np.array(Image.open(BytesIO(image_data)))

    #     print(f'received {req.params}')

    #     resp.status = falcon.HTTP_200
    #     resp.media = {'message': 'Data received successfully'}

class ApiAruco:
    def on_get(self, req, resp):
        # resp.media = json.dumps(api_content_aruco)
        resp.media = api_content_aruco

async def push_height():
    global api_content_esri

    # Size of map is 4993453 bytes

    async with websockets.connect(SOCKET_URI + "push") as websocket:
        await websocket.send(api_content_esri)
        response = await websocket.recv()

async def pull_unity_frame():
    global api_content_unity_png
    async with websockets.connect(SOCKET_URI + "pull") as websocket:
        await websocket.send("Request frame")
        response = await websocket.recv()
        api_content_unity_png = np.array(Image.open(BytesIO(response)))

def main():
    arg_colors = ""
    if arg_colors != "":
        colours = []
        cols = arg_colors.split(",")
        for c in cols:
            colours.append(hex_to_rgb(c))
        print(colours)
        colpoints = np.arange(0, 1, 1/(len(colours) - 1))
        colpoints = np.append(colpoints, [1.0])
        print(colpoints)
        cmap = list(zip(colpoints, colours))
        COLORMAP = create_colormap(cmap)
        COLORMAP_HARD = create_colormap(create_hard_colors(cmap))

    SHOW_LINES = False
    USE_HARD_COLORS = False

    # colours = [(.2, (0,0,128)), (.25, (68, 87, 227)), (.3, (236, 255, 89)),(.4, (67, 138, 65)),(.6, (128,128,128)), (.8, (255, 255, 255))]
    # colours = [(.2, (128,0,0)),(.3, (255,0,0)),(.4, (255,255,255)), (.5, (0,0,255)), (.6,(0, 0, 128))]
    # COLORMAP = create_colormap(color_array=colours)
    # COLORMAP_HARD = create_colormap(create_hard_colors(colours))

    LINE_COLORS = [0,0,0]
    box = Sandbox(refresh=50)
    box.verbosity = 5
    box.init()
    box.on_frame = make_esri_and_send
    box.start(box)

# ---- MAIN ----

from wsgiref.simple_server import make_server
from threading import Thread

app = falcon.App()
app.add_route('/sandbox', ApiUnity())
app.add_route('/aruco', ApiAruco())

class ApiThread(Thread):
    def run(self):
        with make_server('', 8000, app) as httpd:
            print('Serving on port 8000...')

            # Serve until process is killed
            httpd.serve_forever()

ApiThread().start()
main()
